@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2014	PgPool Global Development Group
+ * Copyright (c) 2003-2012	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -56,7 +56,6 @@
 #include "pool_ip.h"
 #include "md5.h"
 #include "pool_stream.h"
-#include "pool_passwd.h"
 
 static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *timeout);
 static StartupPacket *read_startup_packet(POOL_CONNECTION *cp);
@@ -75,8 +74,6 @@ static void init_system_db_connection(void);
 static bool connect_using_existing_connection(POOL_CONNECTION *frontend,
 											  POOL_CONNECTION_POOL *backend,
 											  StartupPacket *sp);
-static void free_persisten_db_connection_memory(POOL_CONNECTION_POOL_SLOT *cp);
-
 /*
  * non 0 means SIGTERM(smart shutdown) or SIGINT(fast shutdown) has arrived
  */
@@ -85,9 +82,6 @@ volatile sig_atomic_t exit_request = 0;
 static int idle;		/* non 0 means this child is in idle state */
 static int accepted = 0;
 
-static int child_inet_fd = 0;
-static int child_unix_fd = 0;
-
 extern int myargc;
 extern char **myargv;
 
@@ -134,11 +128,9 @@ void do_child(int unix_fd, int inet_fd)
 		pool_set_nonblock(inet_fd);
 	}
 #endif
-	if (inet_fd)
-	{
-		child_inet_fd = inet_fd;
-	}
-	child_unix_fd = unix_fd;
+
+	/* Initialize my backend status */
+	pool_initialize_private_backend_status();
 
 	/* Initialize per process context */
 	pool_init_process_context();
@@ -160,15 +152,6 @@ void do_child(int unix_fd, int inet_fd)
 		child_exit(1);
 	}
 
-	/*
-	 * Open pool_passwd in child process.  This is necessary to avoid the
-	 * file descriptor race condition reported in [pgpool-general: 1141].
-	 */
-	if (strcmp("", pool_config->pool_passwd))
-	{
-		pool_reopen_passwd_file();
-	}
-
 	timeout.tv_sec = pool_config->child_life_time;
 	timeout.tv_usec = 0;
 
@@ -181,6 +164,17 @@ void do_child(int unix_fd, int inet_fd)
 		/* pgpool stop request already sent? */
 		check_stop_request();
 
+		/* Check if restart request is set because of failback event
+		 * happend.  If so, exit myself with exit code 1 to be
+		 * restarted by pgpool parent.
+		 */
+		if (pool_get_my_process_info()->need_to_restart)
+		{
+			pool_log("do_child: failback event found. restart myself.");
+			pool_get_my_process_info()->need_to_restart = 0;
+			child_exit(1);
+		}
+
 		accepted = 0;
 
 		/* perform accept() */
@@ -242,7 +236,6 @@ void do_child(int unix_fd, int inet_fd)
 		{
 			pool_debug("SSLRequest from client");
 			pool_ssl_negotiate_serverclient(frontend);
-			pool_free_startup_packet(sp);
 			goto retry_startup;
 		}
 
@@ -276,6 +269,18 @@ void do_child(int unix_fd, int inet_fd)
 		 * Authentication is also done in this step.
 		 */
 
+		/* Check if restart request is set because of failback event
+		 * happend.  If so, close idle connections to backend and make
+		 * a new copy of backend status.
+		 */
+		if (pool_get_my_process_info()->need_to_restart)
+		{
+			pool_log("do_child: failback event found. discard existing connections");
+			pool_get_my_process_info()->need_to_restart = 0;
+			close_idle_connection(0);
+			pool_initialize_private_backend_status();
+		}
+
 		/*
 		 * if there's no connection associated with user and database,
 		 * we need to connect to the backend and send the startup packet.
@@ -358,15 +363,12 @@ void do_child(int unix_fd, int inet_fd)
 			{
 				/* client exits */
 				case POOL_END:
-				case POOL_END_WITH_FRONTEND_ERROR:
 					/*
 					 * do not cache connection if:
-					 * frontend abnormally exits or
 					 * pool_config->connection_cahe == 0 or
 					 * database name is template0, template1, postgres or regression
 					 */
-					if (status == POOL_END_WITH_FRONTEND_ERROR ||
-						pool_config->connection_cache == 0 ||
+					if (pool_config->connection_cache == 0 ||
 						!strcmp(sp->database, "template0") ||
 						!strcmp(sp->database, "template1") ||
 						!strcmp(sp->database, "postgres") ||
@@ -636,11 +638,7 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 	{
 		pool_get_config(get_config_file_name(), RELOAD_CONFIG);
 		if (pool_config->enable_pool_hba)
-		{
 			load_hba(get_hba_file_name());
-			if (strcmp("", pool_config->pool_passwd))
-				pool_reopen_passwd_file();
-		}
 		if (pool_config->parallel_mode)
 			pool_memset_system_db_info(system_db_info->info);
 		got_sighup = 0;
@@ -794,7 +792,9 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 }
 
 /*
-* read startup packet
+* Read startup packet
+*
+* Read the startup packet and parse the contents.
 */
 static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 {
@@ -820,21 +820,19 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 	/* read startup packet length */
 	if (pool_read(cp, &len, sizeof(len)))
 	{
-		pool_error("read_startup_packet: incorrect packet length (%d)", len);
-		pool_free_startup_packet(sp);
-		alarm(0);
-		pool_signal(SIGALRM, SIG_IGN);
 		return NULL;
 	}
 	len = ntohl(len);
 	len -= sizeof(len);
 
-	if (len <= 0 || len >= MAX_STARTUP_PACKET_LENGTH)
+	if (len <= 0)
 	{
 		pool_error("read_startup_packet: incorrect packet length (%d)", len);
+	}
+	else if (len >= MAX_STARTUP_PACKET_LENGTH)
+	{
+		pool_error("read_startup_packet: invalid startup packet");
 		pool_free_startup_packet(sp);
-		alarm(0);
-		pool_signal(SIGALRM, SIG_IGN);
 		return NULL;
 	}
 
@@ -923,6 +921,24 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 						return NULL;
 					}
 				}
+
+				/*
+				 * From 9.0, the start up packet may include
+				 * application name. After receiving such that packet,
+				 * backend sends parameter status of application_name.
+				 * Upon reusing connection to backend, we need to
+				 * emulate this behavior of backend. So we remember
+				 * this and send parameter status packet to frontend
+				 * instead of backend in
+				 * connect_using_existing_connection().
+				 */
+				else if (!strcmp("application_name", p))
+				{
+					p += (strlen(p) + 1);
+					sp->application_name = p;
+					pool_debug("read_startup_packet: application_name: %s", p);
+				}
+
 				p += (strlen(p) + 1);
 			}
 			break;
@@ -957,28 +973,6 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 			return NULL;
 	}
 
-	/* Check a user name was given. */
-	if (sp->major != 1234 &&
-	    (sp->user == NULL || sp->user[0] == '\0'))
-	{
-		pool_send_fatal_message(cp, sp->major, "28000",
-		                        "no PostgreSQL user name specified in startup packet",
-								"",
-								"",
-								__FILE__, __LINE__);
-		pool_error("read_startup_packet: no PostgreSQL user name specified in startup packet");
-		pool_free_startup_packet(sp);
-		alarm(0);
-		pool_signal(SIGALRM, SIG_IGN);
-		return NULL;
-	}
-
-	/* The database defaults to ther user name. */
-	if (sp->database == NULL || sp->database[0] == '\0')
-	{
-		sp->database = strdup(sp->user);
-	}
-
 	pool_debug("Protocol Major: %d Minor: %d database: %s user: %s",
 			   sp->major, sp->minor, sp->database, sp->user);
 	alarm(0);
@@ -987,8 +981,8 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 }
 
 /*
- * send startup packet
- */
+* send startup packet
+*/
 int send_startup_packet(POOL_CONNECTION_POOL_SLOT *cp)
 {
 	int len;
@@ -1001,13 +995,13 @@ int send_startup_packet(POOL_CONNECTION_POOL_SLOT *cp)
 /*
  * Reuse existing connection
  */
-static bool connect_using_existing_connection(POOL_CONNECTION *frontend,
+static bool connect_using_existing_connection(POOL_CONNECTION *frontend, 
 											  POOL_CONNECTION_POOL *backend,
 											  StartupPacket *sp)
 {
 	int i, freed = 0;
 	/*
-	 * save startup packet info
+	 * Save startup packet info
 	 */
 	for (i = 0; i < NUM_BACKENDS; i++)
 	{
@@ -1022,7 +1016,7 @@ static bool connect_using_existing_connection(POOL_CONNECTION *frontend,
 		}
 	}
 
-	/* reuse existing connection to backend */
+	/* Reuse existing connection to backend */
 
 	if (pool_do_reauth(frontend, backend))
 	{
@@ -1033,6 +1027,32 @@ static bool connect_using_existing_connection(POOL_CONNECTION *frontend,
 
 	if (MAJOR(backend) == 3)
 	{
+		char command_buf[1024];
+
+		/* If we have received application_name in the start up
+		 * packet, we send SET command to backend. Also we add or
+		 * replace existing application_name data.
+		 */
+		if (sp->application_name)
+		{
+			snprintf(command_buf, sizeof(command_buf), "SET application_name TO '%s'", sp->application_name);
+
+			for (i=0;i<NUM_BACKENDS;i++)
+			{
+				if (VALID_BACKEND(i))
+					if (do_command(frontend, CONNECTION(backend, i),
+							   command_buf, MAJOR(backend),
+								   MASTER_CONNECTION(backend)->pid,
+								   MASTER_CONNECTION(backend)->key, 0) != POOL_CONTINUE)
+					{
+						pool_error("connect_using_existing_connection: do_command failed. command: %s", command_buf);
+						return false;
+					}
+			}
+
+			pool_add_param(&MASTER(backend)->params, "application_name", sp->application_name);
+		}
+
 		if (send_params(frontend, backend))
 		{
 			pool_close(frontend);
@@ -1041,7 +1061,7 @@ static bool connect_using_existing_connection(POOL_CONNECTION *frontend,
 		}
 	}
 
-	/* send ReadyForQuery to frontend */
+	/* Send ReadyForQuery to frontend */
 	pool_write(frontend, "Z", 1);
 
 	if (MAJOR(backend) == 3)
@@ -1113,7 +1133,7 @@ void cancel_request(CancelPacket *sp)
 		if (!VALID_BACKEND(i))
 			continue;
 
-		if (*(BACKEND_INFO(i).backend_hostname) == '\0')
+		if (*(BACKEND_INFO(i).backend_hostname) == '/')
 			fd = connect_unix_domain_socket(i, TRUE);
 		else
 			fd = connect_inet_domain_socket(i, TRUE);
@@ -1218,14 +1238,6 @@ static RETSIGTYPE die(int sig)
 	switch (sig)
 	{
 		case SIGTERM:	/* smart shutdown */
-			/* Refuse further requests by closing listen socket */
-			if (child_inet_fd)
-			{
-				pool_log("die: close listen socket");
-				close(child_inet_fd);
-			}
-			close(child_unix_fd);
-
 			if (idle == 0)
 			{
 				pool_debug("child receives smart shutdown request but it's not in idle state");
@@ -1389,8 +1401,7 @@ void child_exit(int code)
 	}
 
 	/* let backend know now we are exiting */
-	if (pool_connection_pool)
-		send_frontend_exits();
+	send_frontend_exits();
 
 	exit(code);
 }
@@ -1428,7 +1439,6 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (startup_packet == NULL)
 	{
 		pool_error("make_persistent_db_connection: could not allocate memory");
-		free_persisten_db_connection_memory(cp);
 		return NULL;
 	}
 	memset(startup_packet, 0, sizeof(*startup_packet));
@@ -1437,9 +1447,9 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	/*
 	 * create socket
 	 */
-	if (*hostname == '\0')
+	if (*hostname == '/')
 	{
-		fd = connect_unix_domain_socket_by_port(port, pool_config->backend_socket_dir, TRUE);
+		fd = connect_unix_domain_socket_by_port(port, hostname, TRUE);
 	}
 	else
 	{
@@ -1449,8 +1459,6 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (fd < 0)
 	{
 		pool_error("make_persistent_db_connection: connection to %s(%d) failed", hostname, port);
-		free_persisten_db_connection_memory(cp);
-		free(startup_packet);
 		return NULL;
 	}
 
@@ -1467,9 +1475,6 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (len1 >= (sizeof(startup_packet->data)-len))
 	{
 		pool_error("make_persistent_db_connection: too long user name");
-		pool_close(cp->con);
-		free_persisten_db_connection_memory(cp);
-		free(startup_packet);
 		return NULL;
 	}
 
@@ -1478,9 +1483,6 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (len1 >= (sizeof(startup_packet->data)-len))
 	{
 		pool_error("make_persistent_db_connection: too long user name");
-		pool_close(cp->con);
-		free_persisten_db_connection_memory(cp);
-		free(startup_packet);
 		return NULL;
 	}
 
@@ -1489,9 +1491,6 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (len1 >= (sizeof(startup_packet->data)-len))
 	{
 		pool_error("make_persistent_db_connection: too long database name");
-		pool_close(cp->con);
-		free_persisten_db_connection_memory(cp);
-		free(startup_packet);
 		return NULL;
 	}
 	len += len1;
@@ -1501,9 +1500,6 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (cp->sp == NULL)
 	{
 		pool_error("make_persistent_db_connection: could not allocate memory");
-		pool_close(cp->con);
-		free_persisten_db_connection_memory(cp);
-		free(startup_packet);
 		return NULL;
 	}
 
@@ -1515,16 +1511,12 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (cp->sp->database == NULL)
 	{
 		pool_error("make_persistent_db_connection: could not allocate memory");
-		pool_close(cp->con);
-		free_persisten_db_connection_memory(cp);
 		return NULL;
 	}
 	cp->sp->user = strdup(user);
 	if (cp->sp->user == NULL)
 	{
 		pool_error("make_persistent_db_connection: could not allocate memory");
-		pool_close(cp->con);
-		free_persisten_db_connection_memory(cp);
 		return NULL;
 	}
 
@@ -1535,8 +1527,6 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (status)
 	{
 		pool_error("make_persistent_db_connection: send_startup_packet failed");
-		pool_close(cp->con);
-		free_persisten_db_connection_memory(cp);
 		return NULL;
 	}
 
@@ -1546,37 +1536,12 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (s_do_auth(cp, password))
 	{
 		pool_error("make_persistent_db_connection: s_do_auth failed");
-		pool_close(cp->con);
-		free_persisten_db_connection_memory(cp);
 		return NULL;
 	}
 
 	return cp;
 }
 
-/*
- * Free memory of POOL_CONNECTION_POOL_SLOT.  Should only be used in
- * make_persistent_db_connection and discard_persistent_db_connection.
- */
-void free_persisten_db_connection_memory(POOL_CONNECTION_POOL_SLOT *cp)
-{
-	if (!cp)
-		return;
-	if (!cp->sp)
-	{
-		free(cp);
-		return;
-	}
-	if (cp->sp->startup_packet)
-		free(cp->sp->startup_packet);
-	if (cp->sp->database)
-		free(cp->sp->database);
-	if (cp->sp->user)
-		free(cp->sp->user);
-	free(cp->sp);
-	free(cp);
-}
-
 /*
  * Discard connection and memory allocated by
  * make_persistent_db_connection().
@@ -1604,12 +1569,15 @@ void discard_persistent_db_connection(POOL_CONNECTION_POOL_SLOT *cp)
 	pool_unset_nonblock(cp->con->fd);
 
 	pool_close(cp->con);
-	free_persisten_db_connection_memory(cp);
+	free(cp->sp->startup_packet);
+	free(cp->sp->database);
+	free(cp->sp->user);
+	free(cp->sp);
+	free(cp);
 }
 
 /*
- * Do authentication. Assuming the only caller is
- * *make_persistent_db_connection().
+ * do authentication for cp
  */
 static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 {
@@ -1620,7 +1588,6 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 	char state;
 	char *p;
 	int pid, key;
-	bool keydata_done;
 
 	/*
 	 * read kind expecting 'R' packet (authentication response)
@@ -1752,12 +1719,8 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 	}
 
 	/*
-	 * Read backend key data and wait until Reay for query arriving or
-	 * error happens.
+	 * read pid etc.
 	 */
-
-	keydata_done = false;
-
 	for (;;)
 	{
 		status = pool_read(cp->con, &kind, sizeof(kind));
@@ -1770,7 +1733,6 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 		switch (kind)
 		{
 			case 'K':	/* backend key data */
-				keydata_done = true;
 				pool_debug("s_do_auth: backend key data received");
 
 				/* read message length */
@@ -1800,9 +1762,21 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 					return -1;
 				}
 				cp->key = key;
-				break;
 
-			case 'Z':	/* Ready for query */
+				/* read kind expecting 'Z' (ready for query) */
+				status = pool_read(cp->con, &kind, sizeof(kind));
+				if (status < 0)
+				{
+					pool_error("s_do_auth: error while reading kind");
+					return -1;
+				}
+
+				if (kind != 'Z')
+				{
+					pool_error("s_do_auth: expecting Z got %c", kind);
+					return -1;
+				}
+
 				/* read message length */
 				status = pool_read(cp->con, &length, sizeof(length));
 				if (status < 0)
@@ -1822,22 +1796,16 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 
 				pool_debug("s_do_auth: transaction state: %c", state);
 				cp->con->tstate = state;
-
-				if (!keydata_done)
-				{
-					pool_error("s_do_auth: ready for query arrived before receiving keydata");
-				}
 				return 0;
 				break;
 
 			case 'S':	/* parameter status */
-			case 'N':	/* notice response */
-			case 'E':	/* error response */
-				/* Just throw away data */
+				pool_debug("s_do_auth: parameter status data received");
+
 				status = pool_read(cp->con, &length, sizeof(length));
 				if (status < 0)
 				{
-					pool_error("s_do_auth: error while reading message length. kind:%c", kind);
+					pool_error("s_do_auth: error while reading message length");
 					return -1;
 				}
 
@@ -1850,7 +1818,7 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 				break;
 
 			default:
-				pool_error("s_do_auth: unknown response \"%c\" while processing BackendKeyData",
+				pool_error("s_do_auth: unknown response \"%c\" before processing BackendKeyData",
 						   kind);
 				break;
 		}
@@ -1990,9 +1958,8 @@ static void init_system_db_connection(void)
 {	
 	if (pool_config->parallel_mode || pool_config->enable_query_cache)
 	{
-		int nRet;
-		nRet = system_db_connect();
-		if (nRet || PQstatus(system_db_info->pgconn) != CONNECTION_OK)
+		system_db_connect();
+		if (PQstatus(system_db_info->pgconn) != CONNECTION_OK)
 		{
 			pool_error("Could not make persistent libpq system DB connection");
 		}
@@ -2008,3 +1975,24 @@ static void init_system_db_connection(void)
 		}
 	}
 }
+
+/*
+ * Initialize my backend status and master node id.
+ * We copy the backend status to private area so that
+ * they are not changed while I am alive.
+ */
+void pool_initialize_private_backend_status(void)
+{
+	int i;
+
+	pool_debug("pool_initialize_private_backend_status: initialize backend status");
+
+	for (i=0;i<MAX_NUM_BACKENDS;i++)
+	{
+		private_backend_status[i] = BACKEND_INFO(i).backend_status;
+		/* my_backend_status is referred to by VALID_BACKEND macro. */
+		my_backend_status[i] = &private_backend_status[i];
+	}
+
+	my_master_node_id = REAL_MASTER_NODE_ID;
+}
