@@ -6,7 +6,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2015	PgPool Global Development Group
+ * Copyright (c) 2003-2012	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -44,6 +44,7 @@ typedef enum {
 
 static POOL_DEST send_to_where(Node *node, char *query);
 static void where_to_send_deallocate(POOL_QUERY_CONTEXT *query_context, Node *node);
+static char* remove_read_write(int len, const char *contents, int *rewritten_len);
 
 /*
  * Create and initialize per query session context
@@ -85,18 +86,19 @@ void pool_query_context_destroy(POOL_QUERY_CONTEXT *query_context)
 /*
  * Start query
  */
-void pool_start_query(POOL_QUERY_CONTEXT *query_context, char *query, Node *node)
+void pool_start_query(POOL_QUERY_CONTEXT *query_context, char *query, int len, Node *node)
 {
 	POOL_SESSION_CONTEXT *session_context;
 
 	if (query_context)
 	{
 		session_context = pool_get_session_context();
+		query_context->original_length = len;
+		query_context->rewritten_length = -1;
 		query_context->original_query = query;
 		query_context->rewritten_query = NULL;
 		query_context->parse_tree = node;
-		query_context->virtual_master_node_id = REAL_MASTER_NODE_ID;
-		query_context->num_original_params = -1;
+		query_context->virtual_master_node_id = my_master_node_id;
 		pool_set_query_in_progress();
 		session_context->query_context = query_context;
 	}
@@ -113,7 +115,7 @@ void pool_set_node_to_be_sent(POOL_QUERY_CONTEXT *query_context, int node_id)
 		return;
 	}
 
-	if (node_id < 0 || node_id >= MAX_NUM_BACKENDS)
+	if (node_id < 0 || node_id > MAX_NUM_BACKENDS)
 	{
 		pool_error("pool_set_node_to_be_sent: invalid node id:%d", node_id);
 		return;
@@ -135,7 +137,7 @@ void pool_unset_node_to_be_sent(POOL_QUERY_CONTEXT *query_context, int node_id)
 		return;
 	}
 
-	if (node_id < 0 || node_id >= MAX_NUM_BACKENDS)
+	if (node_id < 0 || node_id > MAX_NUM_BACKENDS)
 	{
 		pool_error("pool_unset_node_to_be_sent: invalid node id:%d", node_id);
 		return;
@@ -166,14 +168,6 @@ void pool_clear_node_to_be_sent(POOL_QUERY_CONTEXT *query_context)
 void pool_setall_node_to_be_sent(POOL_QUERY_CONTEXT *query_context)
 {
 	int i;
-	POOL_SESSION_CONTEXT *sc;
-
-	sc = pool_get_session_context();
-	if (!sc)
-	{
-		pool_error("pool_setall_node_to_be_sent: no session context");
-		return;
-	}
 
 	if (!query_context)
 	{
@@ -183,22 +177,9 @@ void pool_setall_node_to_be_sent(POOL_QUERY_CONTEXT *query_context)
 
 	for (i=0;i<NUM_BACKENDS;i++)
 	{
-		if ((BACKEND_INFO(i)).backend_status == CON_UP ||
-			(BACKEND_INFO((i)).backend_status == CON_CONNECT_WAIT))
-		{
-			/*
-			 * In streaming replication mode, if the node is not
-			 * primary node nor load balance node, there's no point to
-			 * send query.
-			 */
-			if (pool_config->master_slave_mode &&
-				!strcmp(pool_config->master_slave_sub_mode, MODE_STREAMREP) &&
-				i != PRIMARY_NODE_ID && i != sc->load_balance_node_id)
-			{
-				continue;
-			}
+		if (private_backend_status[i] == CON_UP ||
+			(private_backend_status[i] == CON_CONNECT_WAIT))
 			query_context->where_to_send[i] = true;
-		}
 	}
 	return;
 }
@@ -244,7 +225,7 @@ bool pool_is_node_to_be_sent(POOL_QUERY_CONTEXT *query_context, int node_id)
 		return false;
 	}
 
-	if (node_id < 0 || node_id >= MAX_NUM_BACKENDS)
+	if (node_id < 0 || node_id > MAX_NUM_BACKENDS)
 	{
 		pool_error("pool_is_node_to_be_sent: invalid node id:%d", node_id);
 		return false;
@@ -292,7 +273,11 @@ int pool_virtual_master_db_node_id(void)
 	{
 		return sc->query_context->virtual_master_node_id;
 	}
-	return REAL_MASTER_NODE_ID;
+
+	/*
+	 * No query context exists. Returns master node id in private buffer.
+	 */
+	return my_master_node_id;
 }
 
 /*
@@ -343,6 +328,27 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 	{
 		pool_set_node_to_be_sent(query_context, REAL_MASTER_NODE_ID);
 	}
+	else if (MASTER_SLAVE && query_context->is_multi_statement)
+	{
+		/*
+		 * If we are in master/slave mode and we have multi stametemt
+		 * query, we should send it to primary server only. Otherwise
+		 * it is possible to send a write query to standby servers
+		 * because we only use the first element of the multi
+		 * statement query and don't care about the rest.  Typical
+		 * situation where we are bugged by this is, "BEGIN;DELETE
+		 * FROM table;END". Note that from pgpool-II 3.1.0
+		 * transactional statements such as "BEGIN" is unconditionaly
+		 * sent to all nodes(see send_to_where() for more details).
+		 * Someday we might be able to understand all part of multi
+		 * statement queries, but until that day we need this band
+		 * aid.
+		 */
+		if (query_context->is_multi_statement)
+		{
+			pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
+		}
+	}
 	else if (MASTER_SLAVE)
 	{
 		POOL_DEST dest;
@@ -466,7 +472,7 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 	}
 	else if (REPLICATION || PARALLEL_MODE)
 	{
-		if (is_select_query(node, query) && !is_sequence_query(node))
+		if (is_select_query(node, query))
 		{
 			/*
 			 * If a writing function call is used or replicate_select is true,
@@ -538,32 +544,56 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 }
 
 /*
- * Send query and wait for response
- * string:
- *  simple query protocol: a query
- *  extended query protocol: contents of the message
+ * Send simple query and wait for response
  * send_type:
  *  -1: do not send this node_id
  *   0: send to all nodes
  *  >0: send to this node_id
- * kind:
- *  simple query protocol: ""
- *  extended query protocol: a kind
  */
-POOL_STATUS pool_send_and_wait(POOL_QUERY_CONTEXT *query_context, char *string,
-							   int len, int send_type, int node_id, char *kind)
+POOL_STATUS pool_send_and_wait(POOL_QUERY_CONTEXT *query_context,
+							   int send_type, int node_id)
 {
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_CONNECTION *frontend;
 	POOL_CONNECTION_POOL *backend;
 	bool is_commit;
+	bool is_begin_read_write;
 	int i;
-
-	is_commit = is_commit_query(query_context->parse_tree);
+	int len;
+	char *string;
 
 	session_context = pool_get_session_context();
 	frontend = session_context->frontend;
 	backend = session_context->backend;
+	is_commit = is_commit_or_rollback_query(query_context->parse_tree);
+	is_begin_read_write = false;
+	len = 0;
+	string = NULL;
+
+	/*
+	 * If the query is BEGIN READ WRITE or
+	 * BEGIN ... SERIALIZABLE in master/slave mode,
+	 * we send BEGIN to slaves/standbys instead.
+	 * original_query which is BEGIN READ WRITE is sent to primary.
+	 * rewritten_query which is BEGIN is sent to standbys.
+	 */
+	if (pool_need_to_treat_as_if_default_transaction(query_context))
+	{
+		is_begin_read_write = true;
+	}
+	else
+	{
+		if (query_context->rewritten_query)
+		{
+			len = query_context->rewritten_length;
+			string = query_context->rewritten_query;
+		}
+		else
+		{
+			len = query_context->original_length;
+			string = query_context->original_query;
+		}
+	}
 
 	/* Send query */
 	for (i=0;i<NUM_BACKENDS;i++)
@@ -577,8 +607,7 @@ POOL_STATUS pool_send_and_wait(POOL_QUERY_CONTEXT *query_context, char *string,
 
 		/*
 		 * If in master/slave mode, we do not send COMMIT/ABORT to
-		 * slaves/standbys if it's not in I(idle in transaction)
-		 * state.
+		 * slaves/standbys if it's in I(idle) state.
 		 */
 		if (is_commit && MASTER_SLAVE && !IS_MASTER_NODE_ID(i) && TSTATE(backend, i) == 'I')
 		{
@@ -597,43 +626,25 @@ POOL_STATUS pool_send_and_wait(POOL_QUERY_CONTEXT *query_context, char *string,
 			continue;
 		}
 
-		if (*kind == '\0')
+		if (is_begin_read_write)
 		{
-			per_node_statement_log(backend, i, string);
-
-			if (send_simplequery_message(CONNECTION(backend, i), len, string, MAJOR(backend)) != POOL_CONTINUE)
-				return POOL_END;
-		}			
-		else
-		{
-			if (pool_config->log_per_node_statement)
+			if (REAL_PRIMARY_NODE_ID == i)
 			{
-				char msgbuf[QUERY_STRING_BUFFER_LEN];
-
-				if (*kind == 'P' || *kind == 'E')
-				{
-					char *stmt;
-
-					if (query_context->rewritten_query)
-						stmt = query_context->rewritten_query;
-					else
-						stmt = query_context->original_query;
-
-					if (*kind == 'P')
-						snprintf(msgbuf, sizeof(msgbuf), "Parse: %s", stmt);
-					else
-						snprintf(msgbuf, sizeof(msgbuf), "Execute: %s", stmt);
-				}
-				else
-				{
-					snprintf(msgbuf, sizeof(msgbuf), "%c message", *kind);
-				}
-
-				per_node_statement_log(backend, i, msgbuf);
+				len = query_context->original_length;
+				string = query_context->original_query;
 			}
+			else
+			{
+				len = query_context->rewritten_length;
+				string = query_context->rewritten_query;
+			}
+		}
 
-			if (send_extended_protocol_message(backend, i, kind, len, string) != POOL_CONTINUE)
-				return POOL_END;
+		per_node_statement_log(backend, i, string);
+
+		if (send_simplequery_message(CONNECTION(backend, i), len, string, MAJOR(backend)) != POOL_CONTINUE)
+		{
+			return POOL_END;
 		}
 	}
 
@@ -647,6 +658,7 @@ POOL_STATUS pool_send_and_wait(POOL_QUERY_CONTEXT *query_context, char *string,
 		else if (send_type > 0 && i != node_id)
 			continue;
 
+#ifdef NOT_USED
 		/*
 		 * If in master/slave mode, we do not send COMMIT/ABORT to
 		 * slaves/standbys if it's in I(idle) state.
@@ -655,6 +667,15 @@ POOL_STATUS pool_send_and_wait(POOL_QUERY_CONTEXT *query_context, char *string,
 		{
 			continue;
 		}
+#endif
+
+		if (is_begin_read_write)
+		{
+			if(REAL_PRIMARY_NODE_ID == i)
+				string = query_context->original_query;
+			else
+				string = query_context->rewritten_query;
+		}
 
 		if (wait_for_query_response(frontend, CONNECTION(backend, i), MAJOR(backend)) != POOL_CONTINUE)
 		{
@@ -675,9 +696,211 @@ POOL_STATUS pool_send_and_wait(POOL_QUERY_CONTEXT *query_context, char *string,
 		 * occurs. In this case, PostgreSQL does not report
 		 * what statement caused that error and make users
 		 * confused.
-		 */
+		 */		
 		per_node_error_log(backend, i, string, "pool_send_and_wait: Error or notice message from backend: ", true);
 	}
+
+	return POOL_CONTINUE;
+}
+
+/*
+ * Send extended query and wait for response
+ * send_type:
+ *  -1: do not send this node_id
+ *   0: send to all nodes
+ *  >0: send to this node_id
+ */
+POOL_STATUS pool_extended_send_and_wait(POOL_QUERY_CONTEXT *query_context,
+										char *kind, int len, char *contents,
+										int send_type, int node_id)
+{
+	POOL_SESSION_CONTEXT *session_context;
+	POOL_CONNECTION *frontend;
+	POOL_CONNECTION_POOL *backend;
+	bool is_commit;
+	bool is_begin_read_write;
+	int i;
+	int str_len;
+	int rewritten_len;
+	char *str;
+	char *rewritten_begin;
+
+	session_context = pool_get_session_context();
+	frontend = session_context->frontend;
+	backend = session_context->backend;
+	is_commit = is_commit_or_rollback_query(query_context->parse_tree);
+	is_begin_read_write = false;
+	str_len = 0;
+	rewritten_len = 0;
+	str = NULL;
+	rewritten_begin = NULL;
+
+	/*
+	 * If the query is BEGIN READ WRITE or
+	 * BEGIN ... SERIALIZABLE in master/slave mode,
+	 * we send BEGIN to slaves/standbys instead.
+	 * original_query which is BEGIN READ WRITE is sent to primary.
+	 * rewritten_query which is BEGIN is sent to standbys.
+	 */
+	if (pool_need_to_treat_as_if_default_transaction(query_context))
+	{
+		is_begin_read_write = true;
+
+		if (*kind == 'P')
+		{
+			rewritten_begin = remove_read_write(len, contents, &rewritten_len);
+			if (rewritten_begin == NULL)
+				return POOL_END;
+		}
+	}
+	
+	if (!rewritten_begin)
+	{	
+		str_len = len;
+		str = contents;
+	}
+
+	/* Send query */
+	for (i=0;i<NUM_BACKENDS;i++)
+	{
+		if (!VALID_BACKEND(i))
+			continue;
+		else if (send_type < 0 && i == node_id)
+			continue;
+		else if (send_type > 0 && i != node_id)
+			continue;
+
+		/*
+		 * If in reset context, we send COMMIT/ABORT to nodes those
+		 * are not in I(idle) state.  This will ensure that
+		 * transactions are closed.
+		 */
+		if (is_commit && session_context->reset_context && TSTATE(backend, i) == 'I')
+		{
+			pool_unset_node_to_be_sent(query_context, i);
+			continue;
+		}
+
+		if (rewritten_begin)
+		{
+			if (REAL_PRIMARY_NODE_ID == i)
+			{
+				str = contents;
+				str_len = len;
+			}
+			else
+			{
+				str = rewritten_begin;
+				str_len = rewritten_len;
+			}
+		}
+
+		if (pool_config->log_per_node_statement)
+		{
+			char msgbuf[QUERY_STRING_BUFFER_LEN];
+			char *stmt;
+
+			if (*kind == 'P' || *kind == 'E')
+			{
+				if (query_context->rewritten_query)
+				{
+					if (is_begin_read_write)
+					{
+						if (REAL_PRIMARY_NODE_ID == i)
+							stmt = query_context->original_query;
+						else
+							stmt = query_context->rewritten_query;
+					}
+					else
+					{
+						stmt = query_context->rewritten_query;
+					}
+				}
+				else
+				{
+					stmt = query_context->original_query;
+				}
+
+				if (*kind == 'P')
+					snprintf(msgbuf, sizeof(msgbuf), "Parse: %s", stmt);
+				else
+					snprintf(msgbuf, sizeof(msgbuf), "Execute: %s", stmt);
+			}
+			else
+			{
+				snprintf(msgbuf, sizeof(msgbuf), "%c message", *kind);
+			}
+
+			per_node_statement_log(backend, i, msgbuf);
+		}
+
+		if (send_extended_protocol_message(backend, i, kind, str_len, str) != POOL_CONTINUE)
+		{
+			free(rewritten_begin);
+			return POOL_END;
+		}
+	}
+
+	if (!is_begin_read_write)
+	{
+		if (query_context->rewritten_query)
+			str = query_context->rewritten_query;
+		else
+			str = query_context->original_query;
+	}
+
+	/* Wait for response */
+	for (i=0;i<NUM_BACKENDS;i++)
+	{
+		if (!VALID_BACKEND(i))
+			continue;
+		else if (send_type < 0 && i == node_id)
+			continue;
+		else if (send_type > 0 && i != node_id)
+			continue;
+
+		/*
+		 * If in master/slave mode, we do not send COMMIT/ABORT to
+		 * slaves/standbys if it's in I(idle) state.
+		 */
+		if (is_commit && MASTER_SLAVE && !IS_MASTER_NODE_ID(i) && TSTATE(backend, i) == 'I')
+		{
+			continue;
+		}
+
+		if (is_begin_read_write)
+		{
+			if (REAL_PRIMARY_NODE_ID == i)
+				str = query_context->original_query;
+			else
+				str = query_context->rewritten_query;
+		}
+
+		if (wait_for_query_response(frontend, CONNECTION(backend, i), MAJOR(backend)) != POOL_CONTINUE)
+		{
+			/* Cancel current transaction */
+			CancelPacket cancel_packet;
+
+			cancel_packet.protoVersion = htonl(PROTO_CANCEL);
+			cancel_packet.pid = MASTER_CONNECTION(backend)->pid;
+			cancel_packet.key= MASTER_CONNECTION(backend)->key;
+			cancel_request(&cancel_packet);
+
+			free(rewritten_begin);
+			return POOL_END;
+		}
+
+		/*
+		 * Check if some error detected.  If so, emit
+		 * log. This is usefull when invalid encoding error
+		 * occurs. In this case, PostgreSQL does not report
+		 * what statement caused that error and make users
+		 * confused.
+		 */		
+		per_node_error_log(backend, i, str, "pool_send_and_wait: Error or notice message from backend: ", true);
+	}
+
+	free(rewritten_begin);
 	return POOL_CONTINUE;
 }
 
@@ -815,12 +1038,6 @@ static POOL_DEST send_to_where(Node *node, char *query)
 			/* SELECT INTO or SELECT FOR SHARE or UPDATE ? */
 			if (pool_has_insertinto_or_locking_clause(node))
 				return POOL_PRIMARY;
-			/*
-			 * SELECT nextval(), setval()
-			 * XXX: We do not search in subquery.
-			 */
-			else if (is_sequence_query(node))
-				return POOL_PRIMARY;
 
 			return POOL_EITHER;
 		}
@@ -846,39 +1063,23 @@ static POOL_DEST send_to_where(Node *node, char *query)
 		 */
 		else if (IsA(node, TransactionStmt))
 		{
-			ListCell   *list_item;
-
 			/*
 			 * Check "BEGIN READ WRITE" "START TRANSACTION READ WRITE"
 			 */
-			if (((TransactionStmt *)node)->kind == TRANS_STMT_BEGIN ||
-				((TransactionStmt *)node)->kind == TRANS_STMT_START)
+			if (is_start_transaction_query(node))
 			{
-				List *options = ((TransactionStmt *)node)->options;
-				foreach(list_item, options)
-				{
-					DefElem *opt = (DefElem *) lfirst(list_item);
-
-					if (!strcmp("transaction_read_only", opt->defname))
-					{
-						bool read_only;
-
-						read_only = ((A_Const *)opt->arg)->val.val.ival;
-						if (!read_only)
-							return POOL_PRIMARY;
-					}
-				}
-				/* Other TRANSACTION start commands are sent to the primary */
-				return POOL_PRIMARY;
+				/* But actually, we send BEGIN to standby if it's
+				   BEGIN READ WRITE or START TRANSACTION READ WRITE */
+				if (is_read_write((TransactionStmt *)node))
+					return POOL_BOTH;
+				/* Other TRANSACTION start commands are sent to both primary
+				   and standby */
+				else
+					return POOL_BOTH;
 			}
-			else if (((TransactionStmt *)node)->kind == TRANS_STMT_SAVEPOINT ||
-					 ((TransactionStmt *)node)->kind == TRANS_STMT_ROLLBACK_TO ||
-					 ((TransactionStmt *)node)->kind == TRANS_STMT_RELEASE)
-			{
-				/* SAVEPOINT related commands are sent to the primary */
-				return POOL_PRIMARY;
-			}
-
+			/* SAVEPOINT related commands are sent to both primary and standby */
+			else if (is_savepoint_query(node))
+				return POOL_BOTH;
 			/*
 			 * 2PC commands
 			 */
@@ -886,7 +1087,7 @@ static POOL_DEST send_to_where(Node *node, char *query)
 				return POOL_PRIMARY;
 			else
 				/* COMMIT etc. */
-				return POOL_PRIMARY;
+				return POOL_BOTH;
 		}
 
 		/*
@@ -926,6 +1127,16 @@ static POOL_DEST send_to_where(Node *node, char *query)
 				return ret;
 			}
 
+			/* SET TRANSACTION ISOLATION LEVEL SERIALIZABLE or
+			 * SET SESSION CHARACTERISTICS AS TRANSACTION ISOLATION LEVEL SERIALIZABLE or
+			 * SET transaction_isolation TO 'serializable'
+			 * SET default_transaction_isolation TO 'serializable'
+			 */
+			else if (is_set_transaction_serializable(node))
+			{
+				return POOL_PRIMARY;
+			}
+
 			/*
 			 * Check "SET TRANSACTION READ WRITE" "SET SESSION
 			 * CHARACTERISTICS AS TRANSACTION READ WRITE"
@@ -1021,10 +1232,11 @@ void where_to_send_deallocate(POOL_QUERY_CONTEXT *query_context, Node *node)
 	DeallocateStmt *d = (DeallocateStmt *)node;
 	POOL_SENT_MESSAGE *msg;
 
-	/* DEALLOCATE ALL? */
+	/* DELLOCATE ALL? */
 	if (d->name == NULL)
 	{
 		pool_setall_node_to_be_sent(query_context);
+		return;
 	}
 	else
 	{
@@ -1036,11 +1248,11 @@ void where_to_send_deallocate(POOL_QUERY_CONTEXT *query_context, Node *node)
 			/* Inherit same map from PREPARE or PARSE */
 			pool_copy_prep_where(msg->query_context->where_to_send,
 								 query_context->where_to_send);
-			return;
 		}
-		/* prepared statement was not found */
-		pool_setall_node_to_be_sent(query_context);
+		return;
 	}
+	/* prepared statement was not found */
+	pool_setall_node_to_be_sent(query_context);
 }
 
 /*
@@ -1064,7 +1276,7 @@ Node *pool_get_parse_tree(void)
 
 /*
  * Returns raw query string for current query.
- * Preconition: the query is in progress state.
+ * Precondition: the query is in progress state.
  */
 char *pool_get_query_string(void)
 {
@@ -1088,7 +1300,7 @@ char *pool_get_query_string(void)
  * SET transaction_isolation TO 'serializable'
  * SET default_transaction_isolation TO 'serializable'
  */
-bool is_set_transaction_serializable(Node *node, char *query)
+bool is_set_transaction_serializable(Node *node)
 {
 	ListCell   *list_item;
 
@@ -1137,6 +1349,104 @@ bool is_set_transaction_serializable(Node *node, char *query)
 	return false;
 }
 
+/*
+ * Returns true if SQL is transaction starting command (START
+ * TRANSACTION or BEGIN)
+ */
+bool is_start_transaction_query(Node *node)
+{
+	TransactionStmt *stmt;
+
+	if (node == NULL || !IsA(node, TransactionStmt))
+		return false;
+
+	stmt = (TransactionStmt *)node;
+	return stmt->kind == TRANS_STMT_START || stmt->kind == TRANS_STMT_BEGIN;
+}
+
+/*
+ * Return true if start transaction query with "READ WRITE" option.
+ */
+bool is_read_write(TransactionStmt *node)
+{
+	ListCell   *list_item;
+
+	List *options = node->options;
+	foreach(list_item, options)
+	{
+		DefElem *opt = (DefElem *) lfirst(list_item);
+
+		if (!strcmp("transaction_read_only", opt->defname))
+		{
+			bool read_only;
+
+			read_only = ((A_Const *)opt->arg)->val.val.ival;
+			if (read_only)
+				return false;	/* TRANSACTION READ ONLY */
+			else
+				/*
+				 * TRANSACTION READ WRITE specified. This sounds a little bit strange,
+				 * but actually the parse code works in the way.
+				 */
+				return true;
+		}
+	}
+
+	/*
+	 * No TRANSACTION READ ONLY/READ WRITE clause specified.
+	 */
+	return false;
+}
+
+/*
+ * Return true if start transaction query with "SERIALIZABLE" option.
+ */
+bool is_serializable(TransactionStmt *node)
+{
+	ListCell   *list_item;
+
+	List *options = node->options;
+	foreach(list_item, options)
+	{
+		DefElem *opt = (DefElem *) lfirst(list_item);
+
+		if (!strcmp("transaction_isolation", opt->defname) &&
+			IsA(opt->arg, A_Const) &&
+			((A_Const *)opt->arg)->val.type == T_String &&
+			!strcmp("serializable", ((A_Const *)opt->arg)->val.val.str))
+				return true;
+	}
+	return false;
+}
+
+/*
+ * If the query is BEGIN READ WRITE or
+ * BEGIN ... SERIALIZABLE in master/slave mode,
+ * we send BEGIN to slaves/standbys instead.
+ * original_query which is BEGIN READ WRITE is sent to primary.
+ * rewritten_query which is BEGIN is sent to standbys.
+ */
+bool pool_need_to_treat_as_if_default_transaction(POOL_QUERY_CONTEXT *query_context)
+{
+	return (MASTER_SLAVE &&
+			is_start_transaction_query(query_context->parse_tree) &&
+			(is_read_write((TransactionStmt *)query_context->parse_tree) ||
+			 is_serializable((TransactionStmt *)query_context->parse_tree)));
+}
+
+/*
+ * Return true if the query is SAVEPOINT related query.
+ */
+bool is_savepoint_query(Node *node)
+{
+	if (((TransactionStmt *)node)->kind == TRANS_STMT_SAVEPOINT ||
+		((TransactionStmt *)node)->kind == TRANS_STMT_ROLLBACK_TO ||
+		((TransactionStmt *)node)->kind == TRANS_STMT_RELEASE)
+		return true;
+
+	return false;
+}
+
 /*
  * Return true if the query is 2PC transaction query.
  */
@@ -1240,77 +1550,3 @@ char* remove_read_write(int len, const char* contents, int *rewritten_len)
 
 	return rewritten_contents;
 }
-
-/*
- * Return true if one of followings is true
- *
- * SET transaction_read_only TO on
- * SET TRANSACTION READ ONLY
- * SET TRANSACTION CHARACTERISTICS AS TRANSACTION READ ONLY
- *
- * Note that if the node is not a variable statement, returns false.
- */
-bool pool_is_transaction_read_only(Node *node)
-{
-	ListCell   *list_item;
-	bool ret = false;
-
-	if (!IsA(node, VariableSetStmt))
-		return ret;
-
-	/*
-	 * SET transaction_read_only TO on
-	 */
-	if (((VariableSetStmt *)node)->kind == VAR_SET_VALUE &&
-		!strcmp(((VariableSetStmt *)node)->name, "transaction_read_only"))
-	{
-		List *options = ((VariableSetStmt *)node)->args;
-		foreach(list_item, options)
-		{
-			A_Const *v = (A_Const *)lfirst(list_item);
-
-			switch (v->val.type)
-			{
-				case T_String:
-					if (!strcasecmp(v->val.val.str, "on") ||
-						!strcasecmp(v->val.val.str, "t") ||
-						!strcasecmp(v->val.val.str, "true"))
-						ret = true;
-					break;
-				case T_Integer:
-					if (v->val.val.ival)
-						ret = true;
-				default:
-					break;
-			}
-		}
-	}
-
-	/*
-	 * SET SESSION CHARACTERISTICS AS TRANSACTION READ ONLY
-	 * SET TRANSACTION READ ONLY
-	 */
-	else if (((VariableSetStmt *)node)->kind == VAR_SET_MULTI &&
-			 (!strcmp(((VariableSetStmt *)node)->name, "TRANSACTION") ||
-			  !strcmp(((VariableSetStmt *)node)->name, "SESSION CHARACTERISTICS")))
-	{
-		List *options = ((VariableSetStmt *)node)->args;
-		foreach(list_item, options)
-		{
-			DefElem *opt = (DefElem *) lfirst(list_item);
-
-			if (!strcmp("transaction_read_only", opt->defname))
-			{
-				bool read_only;
-
-				read_only = ((A_Const *)opt->arg)->val.val.ival;
-				if (read_only)
-				{
-					ret = true;
-					break;
-				}
-			}
-		}
-	}
-	return ret;
-}
